@@ -233,6 +233,9 @@ g_vfs_backend_archive_finalize (GObject *object)
   GVfsBackendArchive *archive = G_VFS_BACKEND_ARCHIVE (object);
 
   g_assert (archive->file == NULL);
+
+  if (G_OBJECT_CLASS (g_vfs_backend_archive_parent_class)->finalize)
+    (*G_OBJECT_CLASS (g_vfs_backend_archive_parent_class)->finalize) (object);
 }
 
 static void
