@@ -294,6 +294,9 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
     if (new_loc == NULL)
       return;
 
+    if (!SOUP_STATUS_IS_REDIRECTION(status))
+      return;
+
    new_uri = soup_uri_new_with_base (soup_message_get_uri (msg), new_loc);
    if (new_uri == NULL)
      {
@@ -305,6 +308,10 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
 
    old_uri = soup_message_get_uri (msg);
 
+   /* copy over username and password to new_uri */
+   soup_uri_set_user(new_uri, old_uri->user);
+   soup_uri_set_password(new_uri, old_uri->password);
+
    /* Check if this is a trailing slash redirect (i.e. /a/b to /a/b/),
     * redirect it right away
     */
@@ -358,15 +365,13 @@ redirect_handler (SoupMessage *msg, gpointer user_data)
 #endif
          /* ELSE:
           *
-          * Three possibilities:
+          * Two possibilities:
           *
-          *   1) This was a non-3xx response that happened to
-          *      have a "Location" header
-          *   2) It's a non-redirecty 3xx response (300, 304,
+          *   1) It's a non-redirecty 3xx response (300, 304,
           *      305, 306)
-          *   3) It's some newly-defined 3xx response (308+)
+          *   2) It's some newly-defined 3xx response (308+)
           *
-          * We ignore all of these cases. In the first two,
+          * We ignore both of these cases. In the first,
           * redirecting would be explicitly wrong, and in the
           * last case, we have no clue if the 3xx response is
           * supposed to be redirecty or non-redirecty. Plus,
@@ -512,6 +517,36 @@ xml_node_iter_get_current (xmlNodeIter *iter)
   return iter->cur_node;
 }
 
+static gint
+http_to_gio_error(guint status_code)
+{
+  switch (status_code)
+    {
+      case SOUP_STATUS_NOT_FOUND:
+        return G_IO_ERROR_NOT_FOUND;
+        break;
+      case SOUP_STATUS_UNAUTHORIZED:
+      case SOUP_STATUS_PAYMENT_REQUIRED:
+      case SOUP_STATUS_FORBIDDEN:
+        return G_IO_ERROR_PERMISSION_DENIED;
+        break;
+      case SOUP_STATUS_REQUEST_TIMEOUT:
+        return G_IO_ERROR_TIMED_OUT;
+        break;
+      case SOUP_STATUS_CANT_RESOLVE:
+        return G_IO_ERROR_HOST_NOT_FOUND;
+        break;
+      case SOUP_STATUS_NOT_IMPLEMENTED:
+        return G_IO_ERROR_NOT_SUPPORTED;
+        break;
+      case SOUP_STATUS_INSUFFICIENT_STORAGE:
+        return G_IO_ERROR_NO_SPACE;
+        break;
+    }
+  
+  return G_IO_ERROR_FAILED;
+}
+
 static xmlDocPtr
 parse_xml (SoupMessage  *msg,
            xmlNodePtr   *root,
@@ -522,7 +557,7 @@ parse_xml (SoupMessage  *msg,
 
   if (!SOUP_STATUS_IS_SUCCESSFUL (msg->status_code))
     {
-      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+      g_set_error (error, G_IO_ERROR, http_to_gio_error (msg->status_code),
                    _("HTTP Error: %s"), msg->reason_phrase);
       return NULL;
     }
