@@ -98,7 +98,7 @@ static DBusHandlerResult peer_to_peer_filter_func  (DBusConnection *conn,
 						    gpointer        data);
 
 
-G_DEFINE_TYPE (GVfsDaemon, g_vfs_daemon, G_TYPE_OBJECT);
+G_DEFINE_TYPE (GVfsDaemon, g_vfs_daemon, G_TYPE_OBJECT)
 
 static void
 registered_path_free (RegisteredPath *data)
@@ -189,10 +189,11 @@ g_vfs_daemon_set_property (GObject         *object,
 			   const GValue    *value,
 			   GParamSpec      *pspec)
 {
+#if 0
   GVfsDaemon *daemon;
   
   daemon = G_VFS_DAEMON (object);
-  
+#endif
   switch (prop_id)
     {
     default:
@@ -207,10 +208,11 @@ g_vfs_daemon_get_property (GObject    *object,
 			   GValue     *value,
 			   GParamSpec *pspec)
 {
+#if 0
   GVfsDaemon *daemon;
   
   daemon = G_VFS_DAEMON (object);
-
+#endif
   switch (prop_id)
     {
     default:
@@ -307,7 +309,7 @@ static void
 daemon_schedule_exit (GVfsDaemon *daemon)
 {
   if (daemon->exit_tag == 0)
-    daemon->exit_tag = g_timeout_add (1000, exit_at_idle, daemon);
+    daemon->exit_tag = g_timeout_add_seconds (1, exit_at_idle, daemon);
 }
 
 static void
@@ -1048,6 +1050,7 @@ g_vfs_daemon_initiate_mount (GVfsDaemon *daemon,
 			  "daemon", daemon,
 			  "object_path", obj_path,
 			  NULL);
+  g_free (obj_path);
 
   g_vfs_daemon_add_job_source (daemon, G_VFS_JOB_SOURCE (backend));
   g_object_unref (backend);
