@@ -31,6 +31,8 @@
 #include <glib/gi18n-lib.h>
 #include <gio/gio.h>
 
+#include <gvfsmountinfo.h>
+
 #include "ghalvolumemonitor.h"
 #include "ghalmount.h"
 #include "ghalvolume.h"
@@ -57,27 +59,22 @@ struct _GHalMount {
   GIcon *override_icon;
   GFile *override_root;
   gboolean cannot_unmount;
-  gboolean searched_for_icon;
 
   HalDevice *device;
   HalDevice *drive_device;
+
+  GIcon *autorun_icon;
+  gboolean searched_for_autorun;
+
+  gchar *xdg_volume_info_name;
+  GIcon *xdg_volume_info_icon;
+  gboolean searched_for_xdg_volume_info;
 };
 
-static GFile *
-_g_find_file_insensitive_finish (GFile        *parent,
-                                 GAsyncResult *result,
-                                 GError      **error);
-
-static void
-_g_find_file_insensitive_async (GFile              *parent,
-                                const gchar        *name,
-                                GCancellable       *cancellable,
-                                GAsyncReadyCallback callback,
-                                gpointer            user_data);
-
-
 static GFile *get_root (GHalMount *hal_mount);
 
+static void update_from_hal (GHalMount *m, gboolean emit_changed);
+
 static void g_hal_mount_mount_iface_init (GMountIface *iface);
 
 G_DEFINE_TYPE_EXTENDED (GHalMount, g_hal_mount, G_TYPE_OBJECT, 0,
@@ -114,6 +111,13 @@ g_hal_mount_finalize (GObject *object)
   if (mount->override_root != NULL)
     g_object_unref (mount->override_root);
 
+  if (mount->autorun_icon != NULL)
+    g_object_unref (mount->autorun_icon);
+
+  g_free (mount->xdg_volume_info_name);
+  if (mount->xdg_volume_info_icon != NULL)
+    g_object_unref (mount->xdg_volume_info_icon);
+
   if (mount->volume_monitor != NULL)
     g_object_remove_weak_pointer (G_OBJECT (mount->volume_monitor), (gpointer) &(mount->volume_monitor));
   
@@ -147,144 +151,6 @@ changed_in_idle (gpointer data)
   return FALSE;
 }
 
-typedef struct _MountIconSearchData
-{
-  GHalMount *mount;
-  GFile *root;
-} MountIconSearchData;
-
-static void
-clear_icon_search_data (MountIconSearchData *data)
-{
-  if (data->mount)
-    g_object_unref (data->mount);
-  if (data->root)
-    g_object_unref (data->root);
-  g_free (data);
-}
-
-static void
-on_icon_file_located (GObject *source_object, GAsyncResult *res,
-                      gpointer user_data)
-{
-  GFile *icon_file;
-  GIcon *icon;
-  MountIconSearchData *data = (MountIconSearchData *) (user_data);
-  
-  icon_file = _g_find_file_insensitive_finish (G_FILE (source_object),
-                                               res, NULL);
-  
-  /* TODO: check if the file actually exists? */
-
-  icon = g_file_icon_new (icon_file);
-  g_object_unref (icon_file);
-
-  g_hal_mount_override_icon (data->mount, icon);
-  g_object_unref (icon);
-  
-  clear_icon_search_data (data);
-}
-
-static void
-on_autorun_loaded (GObject *source_object, GAsyncResult *res,
-                   gpointer user_data)
-{
-  gchar *content, *relative_icon_path = NULL;
-  gsize content_length;
-  MountIconSearchData *data = (MountIconSearchData *) (user_data);
-  
-  if (g_file_load_contents_finish (G_FILE (source_object), res, &content,
-                                   &content_length, NULL, NULL))
-    {
-      /* Scan through for an "icon=" line. Can't use GKeyFile,
-       * because .inf files aren't always valid key files
-       **/
-      GRegex *icon_regex;
-      GMatchInfo *match_info;
-      
-      /* [^,] is because sometimes the icon= line
-       * has a comma at the end
-       **/
-      icon_regex = g_regex_new ("icon=([^,\\r\\n]+)",
-                                G_REGEX_CASELESS, 0, NULL);
-      g_regex_match (icon_regex, content, 0,
-                     &match_info);
-      
-      /* Even if there are multiple matches, pick only the
-       * first.
-       **/
-      if (g_match_info_matches (match_info))
-        {
-          gchar *chr;
-          gchar *word = g_match_info_fetch (match_info, 1);
-          
-          /* Replace '\' with '/' */
-          while ((chr = strchr (word, '\\')) != NULL)
-            *chr = '/';
-          
-          /* If the file name's not valid UTF-8,
-           * don't even try to load it
-           **/
-          if (g_utf8_validate (word, -1, NULL))
-            relative_icon_path = word;
-          else
-            g_free (word);
-        }
-      
-      g_match_info_free (match_info);
-      
-      g_regex_unref (icon_regex);
-      g_free (content);
-    }
-
-  /* some autorun.in points to the .exe file for the icon; make sure we avoid using that */
-  if (relative_icon_path && !g_str_has_suffix (relative_icon_path, ".exe"))
-    {
-      _g_find_file_insensitive_async (data->root,
-                                      relative_icon_path,
-                                      NULL, on_icon_file_located,
-                                      data);
-      
-      g_free (relative_icon_path);
-    }
-  else
-    clear_icon_search_data (data);
-}
-
-static void
-on_autorun_located (GObject *source_object, GAsyncResult *res,
-                    gpointer user_data)
-{
-  GFile *autorun_path;
-  MountIconSearchData *data = (MountIconSearchData *) (user_data);
-  
-  autorun_path = _g_find_file_insensitive_finish (G_FILE (source_object),
-                                                  res, NULL);
-  if (autorun_path)
-    g_file_load_contents_async (autorun_path, NULL, on_autorun_loaded, data);
-  else
-    clear_icon_search_data (data);
-  
-  g_object_unref (autorun_path);
-}
-
-static void
-_g_find_mount_icon (GHalMount *m)
-{
-  MountIconSearchData *search_data;
-  
-  m->searched_for_icon = TRUE;
-	
-  search_data = g_new0 (MountIconSearchData, 1);
-  search_data->mount = g_object_ref (m);
-  search_data->root = get_root (m);
-  
-  _g_find_file_insensitive_async (search_data->root,
-                                  "autorun.inf",
-                                  NULL, on_autorun_located,
-                                  search_data);
-}
-
 #define KILOBYTE_FACTOR 1000.0
 #define MEGABYTE_FACTOR (1000.0 * 1000.0)
 #define GIGABYTE_FACTOR (1000.0 * 1000.0 * 1000.0)
@@ -314,6 +180,38 @@ format_size_for_display (guint64 size)
   return str;
 }
 
+static void
+got_autorun_info_cb (GObject      *source_object,
+                     GAsyncResult *res,
+                     gpointer      user_data)
+{
+  GHalMount *mount = G_HAL_MOUNT (user_data);
+
+  mount->autorun_icon = g_vfs_mount_info_query_autorun_info_finish (G_FILE (source_object),
+                                                                    res,
+                                                                    NULL);
+
+  update_from_hal (mount, TRUE);
+
+  g_object_unref (mount);
+}
+
+static void
+got_xdg_volume_info_cb (GObject      *source_object,
+                        GAsyncResult *res,
+                        gpointer      user_data)
+{
+  GHalMount *mount = G_HAL_MOUNT (user_data);
+
+  mount->xdg_volume_info_icon = g_vfs_mount_info_query_xdg_volume_info_finish (G_FILE (source_object),
+                                                                               res,
+                                                                               &(mount->xdg_volume_info_name),
+                                                                               NULL);
+  update_from_hal (mount, TRUE);
+
+  g_object_unref (mount);
+}
+
 static void
 do_update_from_hal (GHalMount *m)
 {
@@ -354,8 +252,8 @@ do_update_from_hal (GHalMount *m)
   volume_disc_is_blank = hal_device_get_property_bool (volume, "volume.disc.is_blank");
   volume_disc_type = hal_device_get_property_string (volume, "volume.disc.type");
   is_audio_player = hal_device_has_capability (drive, "portable_audio_player");
-  icon_from_hal = hal_device_get_property_string (drive, "info.desktop.icon");
-  volume_icon_from_hal = hal_device_get_property_string (volume, "info.desktop.icon");
+  icon_from_hal = hal_device_get_property_string (drive, "storage.icon.drive");
+  volume_icon_from_hal = hal_device_get_property_string (volume, "storage.icon.volume");
   name_from_hal = hal_device_get_property_string (drive, "info.desktop.name");
   volume_name_from_hal = hal_device_get_property_string (volume, "info.desktop.name");
 
@@ -447,7 +345,13 @@ do_update_from_hal (GHalMount *m)
       g_free (size);
     }
 
-  if (m->override_name != NULL)
+  /* order of preference : xdg, override, probed */
+  if (m->xdg_volume_info_name != NULL)
+    {
+      m->name = g_strdup (m->xdg_volume_info_name);
+      g_free (name);
+    }
+  else if (m->override_name != NULL)
     {
       m->name = g_strdup (m->override_name);
       g_free (name);
@@ -455,17 +359,42 @@ do_update_from_hal (GHalMount *m)
   else
     m->name = name;
 
-  if (m->override_icon != NULL)
+  /* order of preference: xdg, autorun, override, probed */
+  if (m->xdg_volume_info_icon != NULL)
+    m->icon = g_object_ref (m->xdg_volume_info_icon);
+  else if (m->autorun_icon != NULL)
+    m->icon = g_object_ref (m->autorun_icon);
+  else if (m->override_icon != NULL)
     m->icon = g_object_ref (m->override_icon);
   else
     m->icon = get_themed_icon_with_fallbacks (icon_name,
                                               icon_name_fallback);
 
-  /* If this is a CD-ROM, begin searching for an icon specified in
-   * autorun.inf.
-  **/
-  if (strcmp (drive_type, "cdrom") == 0 && !m->searched_for_icon)
-    _g_find_mount_icon (m);
+  /* search for .xdg-volume-info */
+  if (!m->searched_for_xdg_volume_info)
+    {
+      GFile *root;
+      root = get_root (m);
+      m->searched_for_xdg_volume_info = TRUE;
+      g_vfs_mount_info_query_xdg_volume_info (root,
+                                              NULL,
+                                              got_xdg_volume_info_cb,
+                                              g_object_ref (m));
+      g_object_unref (root);
+    }
+
+  /* search for autorun.inf */
+  if (!m->searched_for_autorun)
+    {
+      GFile *root;
+      root = get_root (m);
+      m->searched_for_autorun = TRUE;
+      g_vfs_mount_info_query_autorun_info (root,
+                                           NULL,
+                                           got_autorun_info_cb,
+                                           g_object_ref (m));
+      g_object_unref (root);
+    }
 }
 
 
@@ -1272,321 +1201,3 @@ g_hal_mount_mount_iface_init (GMountIface *iface)
   iface->guess_content_type_finish = g_hal_mount_guess_content_type_finish;
   iface->guess_content_type_sync = g_hal_mount_guess_content_type_sync;
 }
-
-#define INSENSITIVE_SEARCH_ITEMS_PER_CALLBACK 100
- 
-static void
-enumerated_children_callback (GObject *source_object, GAsyncResult *res,
-                              gpointer user_data);
- 
-static void
-more_files_callback (GObject *source_object, GAsyncResult *res,
-                     gpointer user_data);
-
-static void
-find_file_insensitive_exists_callback (GObject *source_object,
-                                       GAsyncResult *res,
-                                       gpointer user_data);
-
-typedef struct _InsensitiveFileSearchData
-{
-	GFile *root;
-	gchar *original_path;
-	gchar **split_path;
-	gint index;
-	GFileEnumerator *enumerator;
-	GFile *current_file;
-	
-	GCancellable *cancellable;
-	GAsyncReadyCallback callback;
-	gpointer user_data;
-} InsensitiveFileSearchData;
-
-static void
-_g_find_file_insensitive_async (GFile              *parent,
-                                const gchar        *name,
-                                GCancellable       *cancellable,
-                                GAsyncReadyCallback callback,
-                                gpointer            user_data)
-{
-  InsensitiveFileSearchData *data;
-  GFile *direct_file = g_file_get_child (parent, name);
-  
-  data = g_new0 (InsensitiveFileSearchData, 1);
-  data->cancellable = cancellable;
-  data->callback = callback;
-  data->user_data = user_data;
-  data->root = g_object_ref (parent);
-  data->original_path = g_strdup (name);
-  
-  g_file_query_info_async (direct_file, G_FILE_ATTRIBUTE_STANDARD_TYPE,
-			   G_FILE_QUERY_INFO_NONE, G_PRIORITY_DEFAULT,
-			   cancellable,
-			   find_file_insensitive_exists_callback, data);
-  
-  
-}
-
-static void
-clear_find_file_insensitive_state (InsensitiveFileSearchData *data)
-{
-  if (data->root)
-    g_object_unref (data->root);
-  g_free (data->original_path);
-  if (data->split_path)
-    g_strfreev (data->split_path);
-  if (data->enumerator)
-    g_object_unref (data->enumerator);
-  if (data->current_file)
-    g_object_unref (data->current_file);
-  g_free (data);
-}
-
-static void
-find_file_insensitive_exists_callback (GObject *source_object,
-                                       GAsyncResult *res,
-                                       gpointer user_data)
-{
-  GFileInfo *info;
-  InsensitiveFileSearchData *data = (InsensitiveFileSearchData *) (user_data);
-  
-  /* The file exists and can be found with the given path, no need to search. */
-  if ((info = g_file_query_info_finish (G_FILE (source_object), res, NULL)))
-    {
-      GSimpleAsyncResult *simple;
-      
-      simple = g_simple_async_result_new (G_OBJECT (data->root),
-                                          data->callback,
-                                          data->user_data,
-                                          _g_find_file_insensitive_async);
-      
-      g_simple_async_result_set_op_res_gpointer (simple, g_object_ref (source_object), g_object_unref);
-      g_simple_async_result_complete_in_idle (simple);
-      g_object_unref (simple);
-      clear_find_file_insensitive_state (data);
-    }
-  
-  else
-    {
-      data->split_path = g_strsplit (data->original_path, G_DIR_SEPARATOR_S, -1);
-      data->index = 0;
-      data->enumerator = NULL;
-      data->current_file = g_object_ref (data->root);
-      
-      /* Skip any empty components due to multiple slashes */
-      while (data->split_path[data->index] != NULL &&
-             *data->split_path[data->index] == 0)
-        data->index++;
-      
-      g_file_enumerate_children_async (data->current_file,
-                                       G_FILE_ATTRIBUTE_STANDARD_NAME,
-                                       0, G_PRIORITY_DEFAULT,
-                                       data->cancellable,
-                                       enumerated_children_callback, data);
-    }
-  
-  g_object_unref (source_object);
-}
-
-static void
-enumerated_children_callback (GObject *source_object, GAsyncResult *res,
-                              gpointer user_data)
-{
-  GFileEnumerator *enumerator;
-  InsensitiveFileSearchData *data = (InsensitiveFileSearchData *) (user_data);
-  
-  enumerator = g_file_enumerate_children_finish (G_FILE (source_object),
-                                                 res, NULL);
-  
-  if (enumerator == NULL)
-    {
-      GSimpleAsyncResult *simple;
-      GFile *file;
-      
-      simple = g_simple_async_result_new (G_OBJECT (data->root),
-                                          data->callback,
-                                          data->user_data,
-                                          _g_find_file_insensitive_async);
-      
-      file = g_file_get_child (data->root, data->original_path);
-      
-      g_simple_async_result_set_op_res_gpointer (simple, g_object_ref (file), g_object_unref);
-      g_simple_async_result_complete_in_idle (simple);
-      g_object_unref (simple);
-      clear_find_file_insensitive_state (data);
-      return;
-    }
-  
-  data->enumerator = enumerator;
-  g_file_enumerator_next_files_async (enumerator,
-                                      INSENSITIVE_SEARCH_ITEMS_PER_CALLBACK,
-                                      G_PRIORITY_DEFAULT,
-                                      data->cancellable,
-                                      more_files_callback,
-                                      data);
-}
-
-static void
-more_files_callback (GObject *source_object, GAsyncResult *res,
-                     gpointer user_data)
-{
-  InsensitiveFileSearchData *data = (InsensitiveFileSearchData *) (user_data);
-  GList *files, *l;
-  gchar *filename = NULL, *component, *case_folded_name,
-    *name_collation_key;
-  gboolean end_of_files, is_utf8;
-  
-  files = g_file_enumerator_next_files_finish (data->enumerator,
-                                               res, NULL);
-  
-  end_of_files = files == NULL;
-  
-  component = data->split_path[data->index];
-  g_return_if_fail (component != NULL);
-  
-  is_utf8 = (g_utf8_validate (component, -1, NULL));
-  if (is_utf8)
-    {
-      case_folded_name = g_utf8_casefold (component, -1);
-      name_collation_key = g_utf8_collate_key (case_folded_name, -1);
-      g_free (case_folded_name);
-    }
-  
-  else
-    {
-      name_collation_key = g_ascii_strdown (component, -1);
-    }
-  
-  for (l = files; l != NULL; l = l->next)
-    {
-      GFileInfo *info;
-      const gchar *this_name;
-      gchar *key;
-      
-      info = l->data;
-      this_name = g_file_info_get_name (info);
-      
-      if (is_utf8 && g_utf8_validate (this_name, -1, NULL))
-        {
-          gchar *case_folded;
-          case_folded = g_utf8_casefold (this_name, -1);
-          key = g_utf8_collate_key (case_folded, -1);
-          g_free (case_folded);
-        }
-      else
-        {
-          key = g_ascii_strdown (this_name, -1);
-        }
-      
-      if (strcmp (key, name_collation_key) == 0)
-          filename = g_strdup (this_name);
-      g_free (key);
-      
-      if (filename)
-        break;
-    }
-
-  g_list_foreach (files, (GFunc)g_object_unref, NULL);
-  g_list_free (files);
-  g_free (name_collation_key);
-  
-  if (filename)
-    {
-      GFile *next_file;
-      
-      g_file_enumerator_close_async (data->enumerator,
-                                     G_PRIORITY_DEFAULT,
-                                     data->cancellable,
-                                     NULL, NULL);
-      g_object_unref (data->enumerator);
-      data->enumerator = NULL;
-      
-      /* Set the current file and continue searching */
-      next_file = g_file_get_child (data->current_file, filename);
-      g_free (filename);
-      g_object_unref (data->current_file);
-      data->current_file = next_file;
-      
-      data->index++;
-      /* Skip any empty components due to multiple slashes */
-      while (data->split_path[data->index] != NULL &&
-             *data->split_path[data->index] == 0)
-        data->index++;
-      
-      if (data->split_path[data->index] == NULL)
-       {
-          /* Search is complete, file was found */
-          GSimpleAsyncResult *simple;
-          
-          simple = g_simple_async_result_new (G_OBJECT (data->root),
-                                              data->callback,
-                                              data->user_data,
-                                              _g_find_file_insensitive_async);
-          
-          g_simple_async_result_set_op_res_gpointer (simple, g_object_ref (data->current_file), g_object_unref);
-          g_simple_async_result_complete_in_idle (simple);
-          g_object_unref (simple);
-          clear_find_file_insensitive_state (data);
-          return;
-        }
-      
-      /* Continue searching down the tree */
-      g_file_enumerate_children_async (data->current_file,
-                                       G_FILE_ATTRIBUTE_STANDARD_NAME,
-                                       0, G_PRIORITY_DEFAULT,
-                                       data->cancellable,
-                                       enumerated_children_callback,
-                                       data);
-      return;
-    }
-  
-  if (end_of_files)
-    {
-      /* Could not find the given file, abort the search */
-      GSimpleAsyncResult *simple;
-      GFile *file;
-      
-      g_object_unref (data->enumerator);
-      data->enumerator = NULL;
-      
-      simple = g_simple_async_result_new (G_OBJECT (data->root),
-                                          data->callback,
-                                          data->user_data,
-                                          _g_find_file_insensitive_async);
-      
-      file = g_file_get_child (data->root, data->original_path);
-      g_simple_async_result_set_op_res_gpointer (simple, file, g_object_unref);
-      g_simple_async_result_complete_in_idle (simple);
-      g_object_unref (simple);
-      clear_find_file_insensitive_state (data);
-      return;
-    }
-  
-  /* Continue enumerating */
-  g_file_enumerator_next_files_async (data->enumerator,
-                                      INSENSITIVE_SEARCH_ITEMS_PER_CALLBACK,
-                                      G_PRIORITY_DEFAULT,
-                                      data->cancellable,
-                                      more_files_callback,
-                                      data);
-}
-
-static GFile *
-_g_find_file_insensitive_finish (GFile        *parent,
-                                 GAsyncResult *result,
-                                 GError      **error)
-{
-  GSimpleAsyncResult *simple;
-  GFile *file;
-  
-  g_return_val_if_fail (G_IS_SIMPLE_ASYNC_RESULT (result), NULL);
-  
-  simple = G_SIMPLE_ASYNC_RESULT (result);
-  
-  if (g_simple_async_result_propagate_error (simple, error))
-    return NULL;
-  
-  file = G_FILE (g_simple_async_result_get_op_res_gpointer (simple));
-  return g_object_ref (file);
-}
-
