@@ -128,10 +128,11 @@ g_proxy_drive_new (GProxyVolumeMonitor *volume_monitor)
  * boolean              can-poll-for-media
  * boolean              has-media
  * boolean              is-media-removable
+ * boolean              is-media-check-automatic
  * array:string         volume-ids
  * dict:string->string  identifiers
  */
-#define DRIVE_STRUCT_TYPE "(sssbbbasa{ss})"
+#define DRIVE_STRUCT_TYPE "(sssbbbbasa{ss})"
 
 void
 g_proxy_drive_update (GProxyDrive         *drive,
@@ -146,6 +147,7 @@ g_proxy_drive_update (GProxyDrive         *drive,
   dbus_bool_t can_poll_for_media;
   dbus_bool_t has_media;
   dbus_bool_t is_media_removable;
+  dbus_bool_t is_media_check_automatic;
   GPtrArray *volume_ids;
   GHashTable *identifiers;
 
@@ -164,6 +166,8 @@ g_proxy_drive_update (GProxyDrive         *drive,
   dbus_message_iter_next (&iter_struct);
   dbus_message_iter_get_basic (&iter_struct, &is_media_removable);
   dbus_message_iter_next (&iter_struct);
+  dbus_message_iter_get_basic (&iter_struct, &is_media_check_automatic);
+  dbus_message_iter_next (&iter_struct);
 
   volume_ids = g_ptr_array_new ();
   dbus_message_iter_recurse (&iter_struct, &iter_volume_ids_iter);
@@ -210,6 +214,7 @@ g_proxy_drive_update (GProxyDrive         *drive,
   drive->can_poll_for_media = can_poll_for_media;
   drive->has_media = has_media;
   drive->is_media_removable = is_media_removable;
+  drive->is_media_check_automatic = is_media_check_automatic;
   drive->identifiers = identifiers != NULL ? g_hash_table_ref (identifiers) : NULL;
   drive->volume_ids = g_strdupv ((char **) volume_ids->pdata);
 
@@ -397,32 +402,100 @@ g_proxy_drive_get_id (GProxyDrive *drive)
 }
 
 typedef struct {
-  GObject *object;
+  GProxyDrive *drive;
   GAsyncReadyCallback callback;
   gpointer user_data;
+
+  gchar *cancellation_id;
   GCancellable *cancellable;
+  gulong cancelled_handler_id;
 } DBusOp;
 
+static void
+cancel_operation_reply_cb (DBusMessage *reply,
+                           GError      *error,
+                           gpointer     user_data)
+{
+  if (error != NULL)
+    {
+      g_warning ("Error from CancelOperation(): %s", error->message);
+    }
+}
+
+static void
+operation_cancelled (GCancellable *cancellable,
+                     gpointer      user_data)
+{
+  DBusOp *data = user_data;
+  GSimpleAsyncResult *simple;
+  DBusConnection *connection;
+  DBusMessage *message;
+  const char *name;
+
+  G_LOCK (proxy_drive);
+
+  simple = g_simple_async_result_new_error (G_OBJECT (data->drive),
+                                            data->callback,
+                                            data->user_data,
+                                            G_IO_ERROR,
+                                            G_IO_ERROR_CANCELLED,
+                                            _("Operation was cancelled"));
+  g_simple_async_result_complete_in_idle (simple);
+  g_object_unref (simple);
+
+  /* Now tell the remote volume monitor that the op has been cancelled */
+  connection = g_proxy_volume_monitor_get_dbus_connection (data->drive->volume_monitor);
+  name = g_proxy_volume_monitor_get_dbus_name (data->drive->volume_monitor);
+  message = dbus_message_new_method_call (name,
+                                          "/org/gtk/Private/RemoteVolumeMonitor",
+                                          "org.gtk.Private.RemoteVolumeMonitor",
+                                          "CancelOperation");
+  dbus_message_append_args (message,
+                            DBUS_TYPE_STRING,
+                            &(data->cancellation_id),
+                            DBUS_TYPE_INVALID);
+
+  G_UNLOCK (proxy_drive);
+
+  _g_dbus_connection_call_async (connection,
+                                 message,
+                                 -1,
+                                 (GAsyncDBusCallback) cancel_operation_reply_cb,
+                                 NULL);
+  dbus_message_unref (message);
+  dbus_connection_unref (connection);
+}
+
 static void
 eject_cb (DBusMessage *reply,
           GError *error,
           DBusOp *data)
 {
-  GSimpleAsyncResult *simple;
-  if (error != NULL)
-    simple = g_simple_async_result_new_from_error (data->object,
-                                                   data->callback,
-                                                   data->user_data,
-                                                   error);
-  else
-    simple = g_simple_async_result_new (data->object,
-                                        data->callback,
-                                        data->user_data,
-                                        NULL);
-  g_simple_async_result_complete (simple);
-  g_object_unref (simple);
+  if (data->cancelled_handler_id > 0)
+    g_signal_handler_disconnect (data->cancellable, data->cancelled_handler_id);
 
-  g_object_unref (data->object);
+  if (!g_cancellable_is_cancelled (data->cancellable))
+    {
+      GSimpleAsyncResult *simple;
+
+      if (error != NULL)
+        simple = g_simple_async_result_new_from_error (G_OBJECT (data->drive),
+                                                       data->callback,
+                                                       data->user_data,
+                                                       error);
+      else
+        simple = g_simple_async_result_new (G_OBJECT (data->drive),
+                                            data->callback,
+                                            data->user_data,
+                                            NULL);
+      g_simple_async_result_complete (simple);
+      g_object_unref (simple);
+    }
+
+  g_object_unref (data->drive);
+  g_free (data->cancellation_id);
+  if (data->cancellable != NULL)
+    g_object_unref (data->cancellable);
   g_free (data);
 }
 
@@ -442,22 +515,52 @@ g_proxy_drive_eject (GDrive              *drive,
 
   G_LOCK (proxy_drive);
 
+  if (g_cancellable_is_cancelled (cancellable))
+    {
+      GSimpleAsyncResult *simple;
+      simple = g_simple_async_result_new_error (G_OBJECT (drive),
+                                                callback,
+                                                user_data,
+                                                G_IO_ERROR,
+                                                G_IO_ERROR_CANCELLED,
+                                                _("Operation was cancelled"));
+      g_simple_async_result_complete_in_idle (simple);
+      g_object_unref (simple);
+      G_UNLOCK (proxy_drive);
+      goto out;
+    }
+
   data = g_new0 (DBusOp, 1);
-  data->object = g_object_ref (drive);
+  data->drive = g_object_ref (drive);
   data->callback = callback;
   data->user_data = user_data;
-  data->cancellable = cancellable;
+
+  if (cancellable != NULL)
+    {
+      data->cancellation_id = g_strdup_printf ("%p", cancellable);
+      data->cancellable = g_object_ref (cancellable);
+      data->cancelled_handler_id = g_signal_connect (data->cancellable,
+                                                     "cancelled",
+                                                     G_CALLBACK (operation_cancelled),
+                                                     data);
+    }
+  else
+    {
+      data->cancellation_id = g_strdup ("");
+    }
 
   connection = g_proxy_volume_monitor_get_dbus_connection (proxy_drive->volume_monitor);
   name = g_proxy_volume_monitor_get_dbus_name (proxy_drive->volume_monitor);
 
   message = dbus_message_new_method_call (name,
-                                          "/",
+                                          "/org/gtk/Private/RemoteVolumeMonitor",
                                           "org.gtk.Private.RemoteVolumeMonitor",
                                           "DriveEject");
   dbus_message_append_args (message,
                             DBUS_TYPE_STRING,
                             &(proxy_drive->id),
+                            DBUS_TYPE_STRING,
+                            &(data->cancellation_id),
                             DBUS_TYPE_UINT32,
                             &_flags,
                             DBUS_TYPE_INVALID);
@@ -470,6 +573,8 @@ g_proxy_drive_eject (GDrive              *drive,
                                  data);
   dbus_connection_unref (connection);
   dbus_message_unref (message);
+ out:
+  ;
 }
 
 static gboolean
@@ -487,21 +592,30 @@ poll_for_media_cb (DBusMessage *reply,
                    GError *error,
                    DBusOp *data)
 {
-  GSimpleAsyncResult *simple;
-  if (error != NULL)
-    simple = g_simple_async_result_new_from_error (data->object,
-                                                   data->callback,
-                                                   data->user_data,
-                                                   error);
-  else
-    simple = g_simple_async_result_new (data->object,
-                                        data->callback,
-                                        data->user_data,
-                                        NULL);
-  g_simple_async_result_complete (simple);
-  g_object_unref (simple);
+  if (!g_cancellable_is_cancelled (data->cancellable))
+    {
+      GSimpleAsyncResult *simple;
 
-  g_object_unref (data->object);
+      if (error != NULL)
+        simple = g_simple_async_result_new_from_error (G_OBJECT (data->drive),
+                                                       data->callback,
+                                                       data->user_data,
+                                                       error);
+      else
+        simple = g_simple_async_result_new (G_OBJECT (data->drive),
+                                            data->callback,
+                                            data->user_data,
+                                            NULL);
+      g_simple_async_result_complete (simple);
+      g_object_unref (simple);
+    }
+
+  g_object_unref (data->drive);
+  g_free (data->cancellation_id);
+  if (data->cancelled_handler_id > 0)
+    g_signal_handler_disconnect (data->cancellable, data->cancelled_handler_id);
+  if (data->cancellable != NULL)
+    g_object_unref (data->cancellable);
   g_free (data);
 }
 
@@ -519,22 +633,52 @@ g_proxy_drive_poll_for_media (GDrive              *drive,
 
   G_LOCK (proxy_drive);
 
+  if (g_cancellable_is_cancelled (cancellable))
+    {
+      GSimpleAsyncResult *simple;
+      simple = g_simple_async_result_new_error (G_OBJECT (drive),
+                                                callback,
+                                                user_data,
+                                                G_IO_ERROR,
+                                                G_IO_ERROR_CANCELLED,
+                                                _("Operation was cancelled"));
+      g_simple_async_result_complete_in_idle (simple);
+      g_object_unref (simple);
+      G_UNLOCK (proxy_drive);
+      goto out;
+    }
+
   data = g_new0 (DBusOp, 1);
-  data->object = g_object_ref (drive);
+  data->drive = g_object_ref (drive);
   data->callback = callback;
   data->user_data = user_data;
-  data->cancellable = cancellable;
+
+  if (cancellable != NULL)
+    {
+      data->cancellation_id = g_strdup_printf ("%p", cancellable);
+      data->cancellable = g_object_ref (cancellable);
+      data->cancelled_handler_id = g_signal_connect (data->cancellable,
+                                                     "cancelled",
+                                                     G_CALLBACK (operation_cancelled),
+                                                     data);
+    }
+  else
+    {
+      data->cancellation_id = g_strdup ("");
+    }
 
   connection = g_proxy_volume_monitor_get_dbus_connection (proxy_drive->volume_monitor);
   name = g_proxy_volume_monitor_get_dbus_name (proxy_drive->volume_monitor);
 
   message = dbus_message_new_method_call (name,
-                                          "/",
+                                          "/org/gtk/Private/RemoteVolumeMonitor",
                                           "org.gtk.Private.RemoteVolumeMonitor",
                                           "DrivePollForMedia");
   dbus_message_append_args (message,
                             DBUS_TYPE_STRING,
                             &(proxy_drive->id),
+                            DBUS_TYPE_STRING,
+                            &(data->cancellation_id),
                             DBUS_TYPE_INVALID);
   G_UNLOCK (proxy_drive);
 
@@ -545,6 +689,8 @@ g_proxy_drive_poll_for_media (GDrive              *drive,
                                  data);
   dbus_connection_unref (connection);
   dbus_message_unref (message);
+ out:
+  ;
 }
 
 static gboolean
